
{*******************************************************}
{                                                       }
{       Turbo Pascal Version 7.0                        }
{       Graphics Vision Unit                            }
{                                                       }
{       Copyright (c) 1994 Sasha Peslyak                }
{*******************************************************}

unit GraphDrv;
{$B-,I-,P+}        {* Remove $P+ directive if you are using TP 6.0 }
interface

{ GraphDrv for Turbo Pascal 7.0. It can be used with TP 6.0 also if
  some changes are done. These places are marked with '{*'.
  Supports loadable graphics drivers for different graphics modes.
  The drivers must be in PGI (Peslyak Graphics Interface) format.

  Where it is written that the identifier is same as in standard
  Borland's units, you can use Ctrl+F1 for more information.

  If you have some skills of using the standard Graph unit, you
  should use LoadDriver/SetGraphMode instead of InitGraph.

  This unit requires Turbo Vision's Memory unit (either TV 1.0 or 2.0). }

   Uses
      Memory, LDrawFNT;

   Type
      PathStr=     String[79];         { Same as in the standard DOS unit }

   Var
{ Graphics procedures. Most of them are same as in the standard
  Graph unit, but parameter C is added. It is used to specify the color. }

      LowLevelSetGraphMode   :procedure;
      { You'd better use SetGraphMode instead of this procedure ! }
      ClearDevice            :procedure (c                 :Integer);

      PutPixel               :procedure (x, y, c           :integer);
      Line                   :procedure (x1, y1, x2, y2, c :integer);
      HorizLine              :procedure (y, x1, x2, c      :integer);
      Bar                    :procedure (x1, y1, x2, y2, c :integer);
      QuickBar               :procedure (x1, y1, x2, y2, c :integer);
      Rectangle              :procedure (x1, y1, x2, y2, c :integer);
      GetPixel               :function  (x, y              :integer) :integer;

      GetImage               :procedure (x1, y1, x2, y2    :integer;
                                         buf               :pointer);
      PutImage               :procedure (x, y              :integer;
                                         buf               :pointer);
      ImageSize              :function  (x1, y1, x2, y2    :integer) :word;

      PutBlock               :procedure (x, y              :integer;
                                         p                 :pointer);
      ConvertBlock           :procedure (p1, p2            :pointer);

      BufferToScreen         :procedure;
      ScreenToBuffer         :procedure;

      ChangeColor            :procedure (x1,y1,x2,y2,c1,c2 :integer);

      TurnBlockX             :procedure (Source, Target    :Pointer);
      TurnBlockY             :procedure (Source, Target    :Pointer);

   Const
      ProcCount=19;
   Var
      ProcPTRS     :array [0..ProcCount-1] of pointer
      absolute LowLevelSetGraphMode;

   Const                     { Put modes (same as in standard Graph unit) }
      NormalPut=   0;
      AndPut=      1;
      OrPut=       2;
      XORPut=      3;

      DriverId=    'PGI'#9;

   Type
      TDriver=
      record
         Id                  :array [1..4 ] of Char;
         Header              :array [1..21] of Char; { Driver's name }
         Copyright           :array [1..41] of Char;
         HeaderSize          :Word;
         VideoSeg, VideoSize :Word;
         SizeX, SizeY        :Word;
         MaxColor            :Byte;
         BufferSeg           :Word;
         AspX, AspY          :Byte;
         PutMode             :Byte;                  { These two fields can }
         LinePattern         :Byte;                  { be changed directly }
         LineBreak           :Boolean;
         Extra               :array [1..16] of byte; { Reserved }
         MouseConvX,
         MouseConvY          :Byte;                  { Shifts for mouse }
                                                     { position conversions }
         PutModeSupport      :Array [NormalPut..XORPut] of Boolean;

         ProcCount           :Byte;
         ProcOfs             :array [0..ProcCount-1] of Word;
      end;

      TDriverOutput=         (ToScreen, ToBuffer);

   Var
      DriverPTR              :^TDriver;
      DriverSize             :Word;
      DriverVideoSeg         :Word;
      BufferPTR              :Pointer;

   Const
      TextModeNumber         :Word=    $FFFF;
      { Used by SetGraphMode/RestoreTextMode and CloseGraph }

   Const           { Default color palette. Same as in CRT and Graph units }
      Black     = 0;  DarkGray=      8;
      Blue      = 1;  LightBlue=     9;
      Green     = 2;  LightGreen=   10;
      Cyan      = 3;  LightCyan=    11;
      Red       = 4;  LightRed=     12;
      Magenta   = 5;  LightMagenta= 13;
      Brown     = 6;  Yellow=       14;
      LightGray = 7;  White=        15;

      gdOk=        0;
      gdNoMem=     1;
      gdIOError=   2;

   function  LoadDriver(Const d        :PathStr) :Integer;
      {* Remove CONST for TP 6.0 }
      { Loads the driver into memory. Returnes an error code.
        If the driver is invalid, terminates the program.
        If you enable the extended syntax in your program ($X+ directive),
        you can call it as a procedure }

   procedure FreeDriverMem;
      { Releases the memory used by the driver. Used internally. }

   procedure SetGraphMode;
      { Stores current video mode number and
        sets graphics mode of the currently loaded driver. }

   procedure RestoreTextMode;
      { Restores text mode without removing the driver from memory }
   procedure CloseGraph;
      { Same as in the standard Graph unit. Restores text mode and
        releases memory. Use it at the end of your program. }

   function  Supported(var Proc                 ):Boolean;
      { Determines if the specified procedure or function is
        supported by the currently loaded graphics driver }

   function  PutModeSupported(Mode      :Integer):Boolean;
      { Same as previous, but for put modes }

   procedure GetBufferMem;
   procedure FreeBufferMem;
      { Allocate/release memory for the screen buffer.
        Some drivers don't support buffering. }

   procedure SetDriverOutput(o         :TDriverOutput);
   function  GetDriverOutput           :TDriverOutput;
      { Set/query if graphics is written to screen or buffer }

   function  GetMaxX         :integer;
   function  GetMaxY         :integer;
      { Same as in the standard Graph unit }

   Type
      TPolyLine=
      record
         y, x1, x2 :Integer;
      end;

      TPoly=       array [0..8192] of TPolyLine;

   procedure LFillPoly(var pv;
                       n, c            :Integer);
      { Draws a filled polygon. N - number of points, C - color }

   procedure LDrawPoly(var pv;
                       n, c            :Integer);
      { Same as the previous, but draws only the border }

   procedure Ellipse(x, y,
                     sa, ea,
                     xr, yr,
                     c                 :Integer);
      { Draws the an elliptical arc. X, Y - center,
        SA, EA - start and end angles (in degrees),
        XR, YR - horizontal and vertical radii, C - color }

   procedure Arc    (x, y,
                     sa, ea,
                     r,
                     c                 :Integer);
      { Same as the previous, but draws a circular arc }

   procedure Circle(x, y, r, c         :Integer);
      { Draws a circle }

   procedure EllipseRadius(x, y,
                           a,
                           xr, yr,
                           c           :Integer);

   procedure CircleRadius (x, y,
                           a,
                           r,
                           c           :Integer);
      { These two procedures are used to draw lines with given angle }

   Type
      TTextSettings=
      record
         SizeX, DivX, SizeY, DivY      :Integer;
         Horiz, Vert                   :Word;
      end;

   Const           { These constants are same as in the standard Graph unit }
      LeftText=    0;
      CenterText=  1;
      RightText=   2;
      BottomText=  0;
      TopText=     2;

      DTextSettings          :TTextSettings=
      (SizeX:1 ; DivX:1 ; SizeY:1 ; DivY:1 ; Horiz:LeftText ; Vert:TopText);

   procedure SetTextJustify(h, v                 :Word);
      { Same as in the standard Graph unit }

   procedure SetUserCharSize(mx, dx, my, dy      :Integer);
      { Same as in the standard Graph unit }

   procedure OutTextXY(x, y                      :Integer;
                       Const s                   :String;
                       c                         :Integer);
      {* Remove CONST for TP 6.0 }
      { Writes text S at X,Y with color C using current justify and size.
        Warning: PutMode=XORPut is NOT supported here. }

   procedure OutColoredTextXY(x, y               :Integer;
                              Const s            :String;
                              c1, c2             :Integer);
      {* Remove CONST for TP 6.0 }
      { Same as the previous, but changes the color accordingly '~'
        characters in S }

   procedure SetTextAngle(a                      :Integer);
      { Sets the angle (in degrees) for OutTextXY and OutColoredTextXY }

   function  CStrLen(Const S                     :String)  :Integer;
      {* Remove CONST for TP 6.0 }
      { Same as standard Turbo Vision function }

   Const
      FillBufSize  :Word=    4096;
   { Size of the buffer used by FloodFill in bytes }

   procedure FloodFill(x, y, cf, cb              :Integer);
      { Fills a bounded region with color CF using current put mode.
        As soon as it reaches a point with color CB, it stops filling. }

implementation
   var
      ColoredText  :Boolean;
      HighColor    :Integer;

   procedure GraphNotLoadedMsg; far;
   begin
      writeln('Graphics driver not loaded'); Halt(25);
   end;

   procedure NotSupportedMsg; far;
   begin
      RestoreTextMode;
      writeln('This operation isn''t supported by '+
         'the current graphics driver'); Halt(26);
   end;

   function LoadDriver;
   var
      f            :File;
      InvOffset    :Boolean;
      Offset       :Word;
      fs           :LongInt;
      i            :Integer;
   begin
      LoadDriver:=gdOk;

      FreeDriverMem;

      Assign(f, d); Reset(f, 1);
      fs:=FileSize(f);

      if IOResult<>0 then
      begin
         LoadDriver:=gdIOError; Exit;
      end;

      if fs>$FFE0 then fs:=$FFE0;
      DriverSize:=fs;

      DriverPTR:=MemAllocSeg(DriverSize);
      if DriverPTR=Nil then
      begin
         Close(f); LoadDriver:=gdNoMem; Exit;
      end;

      BlockRead(f, DriverPTR^, DriverSize);
      Close(f);

      if IOResult<>0 then
      begin
         FreeDriverMem; LoadDriver:=gdIOError; Exit;
      end;

      if DriverPTR^.Id<>DriverId then
      begin
         WriteLn('Invalid graphics driver'); Halt(27);
      end;

      InvOffset:=False;
      for i:=0 to ProcCount-1 do
      begin
         Offset:=DriverPTR^.ProcOfs[i];
         if (Offset<>0) and (i<DriverPTR^.ProcCount) then
         begin
            ProcPTRS[i]:=PTR(Seg(DriverPTR^)-$10, Offset);
            if Offset>DriverSize+$100 then InvOffset:=True;
         end
         else
            ProcPTRS[i]:=Addr(NotSupportedMsg);
      end;
      if InvOffset or (DriverPTR^.HeaderSize<>SizeOf(TDriver)-2*ProcCount) then
      begin
         writeln('Incorrect graphics driver version'); Halt(28);
      end;

      DriverVideoSeg:=DriverPTR^.VideoSeg;
   end;

   procedure FreeDriverMem;
   var
      i            :Integer;
   begin
      FreeBufferMem;
      for i:=0 to ProcCount-1 do ProcPTRS[i]:=Addr(GraphNotLoadedMsg);
      if DriverPTR<>NIL then FreeMem(DriverPTR, DriverSize);
      DriverPTR:=NIL; BufferPTR:=NIL;
   end;

   procedure SetGraphMode;
   assembler;
   asm
      cmp  TextModeNumber,0FFFFh
      jne  @@1
      mov  ah,0Fh
      int  10h
      xor  ah,ah
      mov  TextModeNumber,ax

      call dword ptr LowLevelSetGraphMode
@@1:
   end;

   procedure RestoreTextMode;
   assembler;
   asm
      mov  ax,TextModeNumber
      cmp  ax,0FFFFh
      je   @@1
      xor  ah,ah
      int  10h
@@1:
      mov  TextModeNumber,0FFFFh
   end;

   procedure CloseGraph;
   begin
      RestoreTextMode; FreeDriverMem;
   end;

   function Supported;
   begin
      Supported:=(DriverPTR<>nil) and (Pointer(Proc)<>Addr(NotSupportedMsg));
   end;

   function PutModeSupported;
   begin
      PutModeSupported:=(DriverPTR<>nil) and DriverPTR^.PutModeSupport[Mode];
   end;

   procedure SetDriverOutput;
   begin
      if DriverPTR=NIL then GraphNotLoadedMsg;
      if o=ToScreen then DriverPTR^.VideoSeg:=DriverVideoSeg else
         if BufferPTR<>NIL then DriverPTR^.VideoSeg:=DriverPTR^.BufferSeg;
   end;

   function  GetDriverOutput;
   begin
      if (DriverPTR^.VideoSeg=DriverPTR^.BufferSeg) and
         (DriverPTR^.BufferSeg<>DriverVideoSeg)
      then GetDriverOutput:=ToBuffer else GetDriverOutput:=ToScreen;
   end;

   function  GetMaxX;
   begin
      GetMaxX:=DriverPTR^.SizeX-1;
   end;

   function  GetMaxY;
   begin
      GetMaxY:=DriverPTR^.SizeY-1;
   end;

   procedure GetBufferMem;
   begin
      if BufferPTR=NIL then
      begin
         if DriverPTR=NIL then GraphNotLoadedMsg;

         BufferPTR:=MemAllocSeg(DriverPTR^.VideoSize);
         FillChar(BufferPTR^, DriverPTR^.VideoSize, 0);

         DriverPTR^.BufferSeg:=Seg(BufferPTR^);
      end;
   end;

   procedure FreeBufferMem;
   begin
      if (BufferPTR<>NIL) and (DriverPTR<>NIL) then
      begin
         SetDriverOutput(ToScreen);
         FreeMem(BufferPTR, DriverPTR^.VideoSize);
         DriverPTR^.BufferSeg:=DriverVideoSeg;
         BufferPTR:=NIL;
      end;
   end;

   procedure LFillPoly;
   var
      i,y,ly,x1,x2 :Integer;
      p            :TPoly absolute pv;
   begin
      for i:=1 to n-1 do
      begin
         if i=1 then ly:=p[i-1].y else ly:=p[i-1].y+1;
         for y:=ly to p[i].y do
         begin
            if p[i].y=ly then
            begin
               x1:=p[i].x1; x2:=p[i].x2;
            end
            else
            begin
               x1:=p[i-1].x1+(p[i].x1-p[i-1].x1)*(y-ly) div (p[i].y-ly);
               x2:=p[i-1].x2+(p[i].x2-p[i-1].x2)*(y-ly) div (p[i].y-ly);
            end;
            HorizLine(y,x1,x2,c);
         end;
      end;
   end;

   procedure LDrawPoly;
   var
      p            :TPoly absolute pv;
      i            :Integer;
   begin
      HorizLine(p[0].y,p[0].x1,p[0].x2,c);
      for i:=1 to n-1 do
      begin
         Line(p[i-1].x1,p[i-1].y,p[i].x1,p[i].y,c);
         Line(p[i-1].x2,p[i-1].y,p[i].x2,p[i].y,c);
      end;
      HorizLine(p[n-1].y,p[n-1].x1,p[n-1].x2,c);
   end;

   procedure Ellipse;
   var
      i,n,a,cx,cy,lx,ly      :integer;
   begin
      DriverPTR^.LineBreak:=True;
      if xr>yr then n:=yr else n:=xr;
      if n<=5 then n:=n shl 1+1 else inc(n,6);
      for i:=0 to n do
      begin
         a:=sa+longint(i)*(ea-sa) div n;
         cx:=x+longint(QSin(a))*xr div MaxInt;
         cy:=y+longint(QCos(a))*yr div MaxInt;
         if a<>sa then Line(lx,ly,cx,cy,c);
         if (a<>sa) or (i=0) then
         begin
            lx:=cx; ly:=cy;
         end;
      end;
      DriverPTR^.LineBreak:=False;
   end;

   procedure Arc;
   begin
      Ellipse(x,y,sa,ea,r*DriverPTR^.AspX div DriverPTR^.AspY,r,c);
   end;

   procedure Circle;
   begin
      Arc(x,y,0,360,r,c);
   end;

   procedure EllipseRadius;
   begin
      Line(x,y,
         x+longint(QSin(a))*xr div MaxInt,y+longint(QCos(a))*yr div MaxInt,c);
   end;

   procedure CircleRadius;
   begin
      EllipseRadius(x,y,a,r*DriverPTR^.AspX div DriverPTR^.AspY,r,c);
   end;

   procedure SetTextJustify;
   begin
      DTextSettings.Horiz:=h; DTextSettings.Vert:=v;
   end;

   procedure SetUserCharSize;
   begin
      with DTextSettings do
      begin
         SizeX:=mx; DivX:=dx; SizeY:=my; DivY:=dy;
      end;
   end;

   procedure OutTextXY;
   type
      FontPTRArray=     array [0..255] of pointer;
      FontPTRArrayPTR=  ^FontPTRArray;

   var
      LastPutMode  :byte;
      sx,sy,i,stp,
      ct           :integer;
      fp           :PFontPTRType;

   begin
      with DTextSettings do
      begin
         if ColoredText then sx:=CStrLen(s) else sx:=Byte(s[0]);
         sx:=(SizeX shl 3)*sx div DivX; sy:=(SizeY shl 3) div DivY;
         case Horiz of
            RightText:  dec(x,sx);
            CenterText: dec(x,sx shr 1);
         end;
         case Vert of
            BottomText: dec(y,sy);
            CenterText: dec(y,sy shr 1);
         end;
         stp:=(SizeX shl 3) div DivX;


         LastPutMode:=DriverPTR^.PutMode;
         if DriverPTR^.PutMode=XORPut then DriverPTR^.PutMode:=NormalPut;

         ct:=c;
         for i:=1 to byte(s[0]) do
         begin
            if s[i]='~' then
            if ct=c then ct:=HighColor else ct:=c else
            begin
               fp:=GetDrawFPTR(PFontPTR,s[i]);
               if fp<>NIL then
                  LDraw8x8(FontPTRArrayPTR(fp^.fp)^[byte(s[i])-byte(fp^.c1)]^,
                     x,y,ct,SizeX,DivX,SizeY,DivY,Line,Rectangle,Bar);
               inc(x,stp);
            end;
         end;
         DriverPTR^.PutMode:=LastPutMode;
      end;
   end;

   procedure OutColoredTextXY;
   begin
      ColoredText:=True; HighColor:=c2;
      OutTextXY(x, y, s, c1);
      ColoredText:=False;
   end;

   procedure SetTextAngle;
   begin
      SetLDrawAngle(a); LDrawAspX:=DriverPTR^.AspX; LDrawAspY:=DriverPTR^.AspY;
   end;

   function CStrLen; assembler;
   asm
      LES   DI,S
      MOV   CL,ES:[DI]
      INC   DI
      XOR   CH,CH
      MOV   BX,CX
      JCXZ  @@2
      MOV   AL,'~'
      CLD
@@1:  REPNE SCASB
      JNE   @@2
      DEC   BX
      JMP   @@1
@@2:  MOV   AX,BX
   end;

   procedure FloodFill;
   type
      CoordArray=  array [1..$4000] of Integer;
   var
      sx,sy,sxnew,synew,temp,MarkMem   :^CoordArray;
      i,n,nnew,maxx,maxy,MaxIndex      :Integer;
      ArrayMem                         :Word;

      procedure FillPixel(x,y          :Integer);
      begin
         if nnew>=MaxIndex then Exit;
         PutPixel(x,y,cf); inc(nnew); sxnew^[nnew]:=x; synew^[nnew]:=y;
      end;

   begin
      if cf>DriverPtr^.MaxColor then cf:=DriverPtr^.MaxColor;
      if cf=cb then Exit;
      if (FillBufSize<32) or (FillBufSize>MaxAvail) then Exit;

      MaxIndex:=FillBufSize shr 3; ArrayMem:=MaxIndex shl 1;

      Mark(MarkMem);
      GetMem(sx,ArrayMem);    GetMem(sy,ArrayMem);
      GetMem(sxnew,ArrayMem); GetMem(synew,ArrayMem);

      maxx:=DriverPTR^.SizeX-1; maxy:=DriverPTR^.SizeY-1;
      nnew:=0;
      if GetPixel(x,y)=cb then FillPixel(x,y);
      repeat
         n:=nnew; nnew:=0;
         temp:=sx; sx:=sxnew; sxnew:=temp; temp:=sy; sy:=synew; synew:=temp;
         for i:=1 to n do
         begin
            x:=sx^[i]; y:=sy^[i];
            if (x>0) and (GetPixel(x-1,y)=cb) then FillPixel(x-1,y);
            if (x<maxx) and (GetPixel(x+1,y)=cb) then FillPixel(x+1,y);
            if (y>0) and (GetPixel(x,y-1)=cb) then FillPixel(x,y-1);
            if (y<maxy) and (GetPixel(x,y+1)=cb) then FillPixel(x,y+1);
         end;
      until nnew=0;

      Release(MarkMem);
   end;

begin
   DriverPTR:=NIL; FreeDriverMem;
end.
