unit Pictures;
interface
uses
   Objects, GraphDrv, ObjGraph, Rects;

const
   pfStoreBack=    $0001;
   pfBuffered=     $0002;

   psVisible=      $0001;
   psTurnedX=      $0010;
   psTurnedY=      $0020;

type
   TImage=
   object(TObject)

      Image        :^TPoint;
      Size         :Word;

      Flags        :Word;
      Back         :^TPoint;
      BackSize     :Word;

      Origin       :TPoint;
      Pos          :TPoint;
      HotSpot      :TPoint;
      MovingRect   :TmRect;
      State        :Word;

      StreamPos    :LongInt;

      constructor Init(AFlags          :Word);
      destructor  Done; virtual;
      procedure   Free;

      procedure GetFrom(var s          :TStream);

      procedure Draw; virtual;

      procedure GetBounds(var Bounds   :TmRect);

      procedure SetMovingRect(var Rect :TmRect);

      procedure SetHotSpot(NewHotSpot  :TPoint);
      procedure CenterHotSpot;

      procedure Show;
      procedure Hide;
      procedure Redraw;

      procedure MoveTo(NewPos          :TPoint); virtual;
      procedure MoveRel(Delta          :TPoint);

      procedure SetState(AState        :Word;
                         Enable        :Boolean);

   private

      procedure CalcPos;
      procedure GetImage(Buffer        :Pointer);
      procedure PutImage(Buffer        :Pointer);
      function  ImageSize              :Word;

      procedure Turn(Flag              :Word);
   end;

   PPicture=       ^TPicture;
   TPicture=
   object(TImage)

      procedure Draw; virtual;

   end;

implementation

   constructor TImage.Init;
   begin
      Inherited Init;

      MovingRect:=ScreenRect;
      Flags:=AFlags;
      if Flags and pfStoreBack=0 then Flags:=Flags and (not pfBuffered);
   end;

   destructor  TImage.Done;
   begin
      Inherited Done;

      Hide; Free;
   end;

   procedure TImage.Free;
   begin
      if Size<>0 then FreeMem(Image, Size);
      if BackSize<>0 then FreeMem(Back, BackSize);
      Size:=0; BackSize:=0;
   end;

   procedure TImage.GetFrom;
   var
      PictSize     :LongInt;
      NewSize      :Word absolute PictSize;
   begin
      StreamPos:=s.GetPos;
      s.Read(PictSize, 4);
      if PictSize>$FFF0 then Exit;
      if (Size<>0) and (Size<>NewSize) then FreeMem(Image, Size);
      if (Size<>NewSize) and (NewSize>MaxAvail) then Size:=0 else
      begin
         if Size<>NewSize then
         begin
            Size:=NewSize; GetMem(Image, Size);
         end;
         s.Read(Image^, Size);
         State:=State and (not (psTurnedX or psTurnedY));
         if s.Status<>stOk then Size:=0;
      end;
   end;

   procedure TImage.Show;
   var
      NewBackSize  :Word;
   begin
      if (Size<>0) and (State and psVisible=0) then
      begin
         if Flags and pfStoreBack<>0 then
         begin
            NewBackSize:=ImageSize;

            if BackSize<>NewBackSize then
            begin
               if BackSize<>0 then FreeMem(Back, BackSize);

               if NewBackSize>MaxAvail then
               begin
                  BackSize:=0; Exit;
               end;

               BackSize:=NewBackSize; GetMem(Back, BackSize);
            end;

            GetImage(Back);
         end;

         Draw;
         State:=State or psVisible;
      end;
   end;

   procedure TImage.Hide;
   begin
      if BackSize<>0 then
      begin
         PutImage(Back);
         FreeMem(Back, BackSize); BackSize:=0;
         State:=State and (not psVisible);
      end;
   end;

   procedure TImage.Redraw;
   begin
      if State and psVisible<>0 then TImage.MoveTo(Pos);
   end;

   procedure TImage.MoveTo;
   var
      Buffer                 :Pointer;
      BufSize                :Word;
      Bounds                 :TmRect;
      NewOrigin              :TPoint;
   begin
      MovingRect.MovePoint(NewPos);

      NewOrigin:=NewPos;
      Dec(NewOrigin.X, HotSpot.X); Dec(NewOrigin.Y, HotSpot.Y);
      Bounds.A:=NewOrigin;
      Bounds.B.X:=Bounds.A.X+Image^.X-1; Bounds.B.Y:=Bounds.A.Y+Image^.Y-1;

      if ScreenRect.FContains(Bounds.A) and ScreenRect.FContains(Bounds.B) {and
         MovingRect.FContains(NewPos)} then
      if State and psVisible<>0 then
      begin
         if (Flags and pfBuffered<>0) and (GetDriverOutput=ToScreen) then
         begin
            GetBounds(Bounds);
            with Bounds do
            begin
               if Back^.X>Image^.X then inc(B.X, Back^.X-Image^.X);
               if Back^.Y>Image^.Y then inc(B.Y, Back^.Y-Image^.Y);
               if NewOrigin.X<A.X then A.X:=NewOrigin.X else
                  inc(B.X, NewOrigin.X-A.X);
               if NewOrigin.Y<A.Y then A.Y:=NewOrigin.Y else
                  inc(B.Y, NewOrigin.Y-A.Y);
               BufSize:=ObjGraph.ImageSize(Bounds);
            end;

            if BufSize<=MaxAvail then
            begin
               GetMem(Buffer, BufSize);
               ObjGraph.GetImage(Bounds, Buffer);
               SetDriverOutput(ToBuffer);
               ObjGraph.PutImage(Bounds.A, Buffer);
            end
            else Buffer:=nil;
         end
         else Buffer:=nil;

         Hide;
         Origin:=NewOrigin;
         Show;

         if Buffer<>nil then
         begin
            ObjGraph.GetImage(Bounds, Buffer);
            SetDriverOutput(ToScreen);
            ObjGraph.PutImage(Bounds.A, Buffer);
            FreeMem(Buffer, BufSize);
         end;
      end
      else Origin:=NewOrigin;
      CalcPos;
   end;

   procedure TImage.MoveRel;
   var
      NewPos       :TPoint;
   begin
      if LongInt(Delta)=0 then Exit;
      NewPos.X:=Origin.X+Delta.X+HotSpot.X;
      NewPos.Y:=Origin.Y+Delta.Y+HotSpot.Y;
      MoveTo(NewPos);
   end;

   procedure TImage.SetState;
   begin
      if (AState and psVisible<>0) then
         if Enable then Show else Hide;

      if AState and psTurnedX<>0 then
         if Enable xor (State and psTurnedX<>0) then Turn(psTurnedX);

      if AState and psTurnedY<>0 then
         if Enable xor (State and psTurnedY<>0) then Turn(psTurnedY);
   end;

   procedure TImage.GetImage;
   var
      Bounds       :TmRect;
   begin
      GetBounds(Bounds);
      ObjGraph.GetImage(Bounds, Buffer);
   end;

   procedure TImage.PutImage;
   begin
      ObjGraph.PutImage(Origin, Buffer);
   end;

   function TImage.ImageSize;
   var
      Bounds       :TmRect;
   begin
      GetBounds(Bounds);
      ImageSize:=ObjGraph.ImageSize(Bounds);
   end;

   procedure TImage.Draw;
   begin
      ObjGraph.PutImage(Origin, Image);
   end;

   procedure TImage.GetBounds;
   begin
      Bounds.A:=Origin;
      Bounds.B.X:=Origin.X+Image^.X-1; Bounds.B.Y:=Origin.Y+Image^.Y-1;
   end;

   procedure TImage.SetMovingRect;
   var
      LastState    :Word;
   begin
      if ScreenRect.FContains(Rect.A) and ScreenRect.FContains(Rect.B) then
      begin
         MovingRect:=Rect;
         if not MovingRect.FContains(Pos) then
         begin
            LastState:=State;
            Hide;
            Origin.X:=MovingRect.A.X-HotSpot.X;
            Origin.Y:=MovingRect.A.Y-HotSpot.Y;
            if Origin.X<ScreenRect.A.X then Origin.X:=ScreenRect.A.X;
            if Origin.Y<ScreenRect.A.Y then Origin.Y:=ScreenRect.A.Y;
            CalcPos;
            if LastState and psVisible<>0 then Show;
         end;
      end;
   end;

   procedure TImage.SetHotSpot;
   begin
      HotSpot:=NewHotSpot;
      CalcPos;
   end;

   procedure TImage.CenterHotSpot;
   var
      NewHotSpot   :TPoint;
   begin
      if Image<>nil then
      begin
         NewHotSpot.X:=Image^.X shr 1; NewHotSpot.Y:=Image^.Y shr 1;
         SetHotSpot(NewHotSpot);
      end;
   end;

   procedure TImage.CalcPos;
   begin
      Pos:=Origin;
      Inc(Pos.X, HotSpot.X); Inc(Pos.Y, HotSpot.Y);
   end;

   procedure TImage.Turn;
   var
      Buffer       :^TPoint;
   begin
      if Size<=MaxAvail then
      begin
         GetMem(Buffer, Size);
         if Flag=psTurnedX then GraphDrv.TurnBlockX(Image, Buffer)
                           else GraphDrv.TurnBlockY(Image, Buffer);
         Move(Buffer^, Image^, Size);
         FreeMem(Buffer, Size);
         State:=(State and (not Flag)) or ((not (State and Flag)) and Flag);
      end;
   end;

   procedure TPicture.Draw;
   begin
      ObjGraph.PutBlock(Origin, Image);
   end;

end.
